#include "emit.hh"

/* help functions */
char *ExpandEscapers(char *tex);
const char *FormatForTargetInstruction(TargetInstruction *tinst);
std::string LookupParameterForSlot(Schedule::slot_t &slot,
				   const std::string &parameter);

bool Emitter::always_mnemonic = false;
bool Emitter::debug_mode = false;

void Emitter::emit_list(std::vector<std::vector<Schedule::slot_t> > &s)
{
  std::vector<std::vector<Schedule::slot_t> >::iterator iter;
  for(iter=s.begin();
      iter!=s.end();
      iter++)
  {
    this->emit_row(*iter);
  }
}

void Emitter::emit_row(std::vector<Schedule::slot_t> &r)
{ 
  error_has_accured_on_current_row = false;
  rowerrors.clear();
  
  row_count++;
  if( Emitter::debug_mode ) std::cout << row_count << ".\t";
  
  int i;
  // emit prefix
  if( slot_row_prefix != 0 ) std::cout << slot_row_prefix;
  
  for(i=0;i<r.size();i++)
  {
    this->emit_slot(r[i]);
    if( i+1 < r.size() )
    {
      if( !implicit_nops ||
	  r[i+1].second.first != NULL )
      {
	if( slot_separator ) std::cout << slot_separator;
      }
      else break;
    }
  }
  
  if( slot_row_suffix != 0 ) std::cout << slot_row_suffix;
  
  if( error_has_accured_on_current_row )
    std::cout << rowerrors.str() << std::endl;
}
			   
/* hantera \n och \t */
void Emitter::emit_slot(Schedule::slot_t &slot)
{
  Schedule::inst_instance_t iit = slot.second;
  if( iit.first == 0 )
  {
    std::cout << NOP_string;    return;
  }
  const char *format = FormatForTargetInstruction(iit.first);
  
  std::string inbuffer = std::string(format);
  std::istringstream input(inbuffer,std::istringstream::in);
  
  std::ostringstream output;

  if( Emitter::debug_mode )
  {
    const char *temp = iit.first->get_mnemonic();
    int l = strlen(temp);
    std::cout << temp << " ";
    int i;
    for(i=0;i<30-l;i++) std::cout << " ";
  }
  
  while(true)
  {
    char c;
    input.get(c);
    if(input.eof()) break;
    if( c == '{' )
    {
      std::ostringstream vs;
      while(1)
      {
	input.get(c);
	if( c == '}' )
	{
	  break;
	}
	else
	  vs << c;
      }
      std::string substitute = LookupParameterForSlot(slot,vs.str());
      if( substitute != "" )
      {
	output << substitute;
      }
      else
      {
	error_has_accured_on_current_row = true;
	rowerrors << "FORMAT: "
		  << format 
		  << " : {"
		  << vs.str()
		  << "} is undefined" << std::endl;
	
	output << "{" << vs.str() << "}";
      }
    }
    else if( c == '\\' )
    {
      input.get(c);
      if( c == 'n' )
	output << '\n';
      else if( c == 't' )
	output << '\t';
      else if( c == 't' )
	output << '\\';
      else if( c == '{' )
	output << '{';
      else if( c == '}' )
	output << '}';
      else
      {
	output << '\\' << c;
	error_has_accured_on_current_row = true;
	rowerrors << "Unknown escape sequence \\" << c << std::endl;
      }
    }
    else
    {
      output << c;
    }
  }
  //emit assembler code for the slot
  std::cout << output.str();
}

Emitter::Emitter(Architecture *architecture)
{
  slot_row_prefix  = "";
  slot_row_suffix  = "\n";
  slot_separator   = " || ";
  NOP_string       = "NOP";
  implicit_nops    = false;
  
  row_count = 0;
  
  if( architecture )
  {
    char *temp;
    temp = architecture->get_defined_string("slot_row_prefix");
    if( temp ) slot_row_prefix = temp;
    temp = architecture->get_defined_string("slot_row_suffix");
    if( temp ) slot_row_suffix = temp;
    temp = architecture->get_defined_string("slot_separator");
    if( temp ) slot_separator = temp;
    temp = architecture->get_defined_string("nop_instruction");
    if( temp ) NOP_string = temp;
    temp = architecture->get_defined_string("implicit_nops");
    if( temp && (std::string(temp) == "NO" ||
		 std::string(temp) == "No" ||
		 std::string(temp) == "no" ))
    {
      implicit_nops = false;
    }
    else if( temp && (std::string(temp) == "YES" ||
		      std::string(temp) == "Yes" ||
		      std::string(temp) == "yes" ))
    {
      implicit_nops = true;
    }
  }
  //process the strings to handle \t, \\, and \n
  slot_row_prefix = ExpandEscapers(slot_row_prefix);
  slot_row_suffix = ExpandEscapers(slot_row_suffix);
  slot_separator  = ExpandEscapers(slot_separator);
}


Emitter::~Emitter()
{
  free(slot_row_prefix);
  free(slot_row_suffix);
  free(slot_separator);
}

void Emitter::emit(Schedule *s)
{
  this->emit_list(s->s);
  std::cout << std::endl;
}

//----------------------------------------------------------------------

/** Creates a copy of <i>text</i> where <b>\t</b>,<b>\n</b> are
 *  replaced by the appropriate characters. The result is malloc:ed
 *  and it is the reponsibility of the caller to free it.
 */
char *ExpandEscapers(char *text)
{
  char *result;
  int i=0,j=0;
  
  result = (char *)malloc(sizeof(char)*strlen(text)+1); //a little to much
  for(;i<strlen(text);i++,j++)
  {
    if( text[i] == '\\' )
    {
      i++; //always ok since last is \0 for C strings
      if( text[i] == 'n' )
	result[j] = '\n';
      else if( text[i] == 't' )
	result[j] = '\t';
      else if( text[i] == '\\' )
	result[j] = '\\';
      else
      {
	result[j] = '?';
	std::cerr << "Unknown escape sequence \\" << text[i] << std::endl;
      }
    }
    else
    {
      result[j] = text[i];
    }
  }
  result[j] = '\0';
  
  return result;
}

const char *FormatForTargetInstruction(TargetInstruction *tinst)
{
  if( tinst )
  {
    if( Emitter::always_mnemonic )
      return tinst->get_mnemonic();
    if( tinst->format &&  std::string(tinst->format) != "" )
      return tinst->format;
    else
      return tinst->get_mnemonic();
  }
  else
    return "no tinst";
}

bool string_has_suffix(std::string source,std::string suffix)
{
  if( source.length() > suffix.length() )
  {
    int sl = source.length();
    if( source.substr(sl-suffix.length(),suffix.length()) == suffix )
      return true;
  }
  return false;
}
std::string string_without_suffix(std::string source,std::string suffix)
{
  std::string temp = source;
  return source.erase(source.length()-suffix.length(),suffix.length());
}
/**
 *  Those that are special are
 *
 * id  - mnemonic
 *
 * suffixes:
 *
 * .SYM0.NAME
 * .SYM0.X.NAME
 * .SYM0.X.OFFSET
 */
std::string LookupParameterForSlot(Schedule::slot_t &slot,
				   const std::string &parameter)
{
  std::map<char*,std::pair<char*,lcc::Node>,ltstr> & mapping = slot.second.second;
  if( "id" == parameter )
  {
    return std::string(slot.second.first->get_mnemonic());
  }

#define accessm(spec, access) \
  else if( string_has_suffix(parameter,#spec) )\
  {\
    std::string vname = string_without_suffix(parameter,#spec);\
    if( mapping.count((char*)vname.c_str()) )\
    {\
      lcc::Node node = mapping[(char*)vname.c_str()].second;\
      if( node )\
      {\
	std::ostringstream out;\
	out << node->access;\
	return out.str();\
      }\
    }\
    return std::string("");\
  }

#include "accessm.hh"

#undef accessm

  else if( mapping.count((char *)parameter.c_str()) )
  {
    char *tmp_str = (char *)parameter.c_str();
    char *map = mapping[tmp_str].first;
    char *men = (char *)slot.second.first->get_mnemonic();
    return std::string(mapping[(char *)parameter.c_str()].first);
  }
  return std::string("");
}

