#include "epnode.hh"

#define max(a,b) (((a)>(b))?(a):(b))

ENode::ENode(NodeAttr * _node_attr, TimeProfile *_time_profile,
	     Architecture *_archi)
  : node_attr(_node_attr), e(0.0), reference_time(0)
{
  time_profile = new TimeProfile(*_time_profile);
  s = new Schedule(_archi->get_omega());
  space_profile = new SpaceProfile(*_archi);
  power_profile = new PowerProfile(_archi->get_omega(), _archi->get_funits());
}

ENode::ENode(const ENode & _enode)
  : e(_enode.e), reference_time(_enode.reference_time),
    node_attr(_enode.node_attr)
{
  time_profile = new TimeProfile(*(_enode.time_profile));
  s = new Schedule(*(_enode.s));
  space_profile = new SpaceProfile(*(_enode.space_profile));
  power_profile = new PowerProfile(*(_enode.power_profile));
}

ENode::~ENode()
{
  delete time_profile;
  delete s;
  delete space_profile;
  delete power_profile;
}

ENode & ENode::operator=(const ENode & _enode)
{
  node_attr = _enode.node_attr;
  reference_time = _enode.reference_time;
  e = _enode.e;
  time_profile = new TimeProfile(*(_enode.time_profile));
  s = new Schedule(*(_enode.s));
  space_profile = new SpaceProfile(*(_enode.space_profile));
  power_profile = new PowerProfile(*(_enode.power_profile));
  return *this;
}

bool ENode::operator==(const ENode & _enode) const
{
  return ((_enode.node_attr == node_attr) && 
	  (*_enode.space_profile == *space_profile) &&
	  (*_enode.power_profile == *power_profile));
}


std::pair< bool, float >
ENode::insert(const vertex_set_t & _chi,
	      vertex_id & _root, DAG & _dag, 
	      node_map_t & _arg_map, pattern_t & _pattern,
	      TargetInstruction *_target_instruction,
	      Architecture *_architecture, int _offset)
{
  int init_reference_time = reference_time;
  std::map< char *, std::pair< char *, lcc::Node >, ltstr > ops_mapping;
  vertex_info_map::type vim = vertex_info_property(_pattern);
  int root_rc;

  //Advance
  if(_offset > 0) {
    s->advance(_offset);
    time_profile->advance(_offset);
    reference_time += _offset;
  }
  //Arguments
  for(node_map_t::iterator arg_iter = _arg_map.begin(); 
      arg_iter != _arg_map.end(); ++arg_iter) {

    char *arg_id = vim[arg_iter->first].id;
    int residence_class = _target_instruction->get_residence_class(arg_id);
    int earliest_instance = -1;
    std::pair< bool, std::pair< char *, lcc::Node > > ip = 
      space_profile->get_instance(arg_iter->second, residence_class,
				  reference_time, _architecture,
				  &earliest_instance);
    if(!ip.first) {
      if(earliest_instance == -1) {
	return std::make_pair(false, 0.0);
      }
      //Need insert NOPs.
      int NOP_offset = (earliest_instance - reference_time);
      assert(NOP_offset > 0);
      time_profile->advance(NOP_offset);
      s->advance(NOP_offset);
      reference_time += NOP_offset;
    }
    ops_mapping[arg_id] = ip.second;
  }

  const reserv_tab_t & reserv_tab = _target_instruction->get_reserv_tab();
  if(!s->available_slot()) {
    s->advance(1);
    time_profile->advance(1);
    reference_time += 1;
  }
  assert(s->available_slot());
  int tp_offset = time_profile->insert(_chi, reserv_tab, _dag);
  s->advance(tp_offset);
  reference_time += tp_offset;

  //SP
  //This should be a field of reservation table. FIXME resrev_tab.height;
  int height = 0;
  for(int i = 0; i < reserv_tab.size(); i++) {
    height = max(reserv_tab[i].size(), height); 
  }

  std::pair< bool, std::pair< char *, lcc::Node > > p;
  p = space_profile->insert(node_attr->alive, 
			   _root, _dag,
			    (reference_time + height),
			   (root_rc = _target_instruction->get_residence_class("dest1")),
			   _architecture);
  if(!p.first) {
    return std::make_pair(false, 0.0);
  }

  //Destination of the root + mapping
  if(root_rc != STMT_RC) {
    ops_mapping["dest1"] = p.second;
    if(num_vertices(_pattern) == 0 && _arg_map.size() == 0) {
      //transfer (src1 is the only argument - no pattern)
      int t_residence_class = _target_instruction->get_residence_class("src1");
      int earliest_instance = -1;
      std::pair< bool, std::pair< char *, lcc::Node > > ip = 
	space_profile->get_instance(_root, t_residence_class, 
				    reference_time, _architecture, 
				    &earliest_instance);
      if(!ip.first) {
	return std::make_pair(false, 0.0);
      }
      ops_mapping["src1"] = ip.second;
    }
  }

  //Schedule: all mapping is done now
  //Advancing one slot in a time profile does not change its alive set
  bool b_inserted = false;
  int additional_offset = 0;
  tie(b_inserted, additional_offset) = 
    s->insert(_target_instruction, ops_mapping, &_pattern);
  assert(additional_offset == 0);

  assert(s->size() == reference_time + 1);

  //Power Profile - contribution from init_ref_time to ref_time
  int f_units = _architecture->get_funits();
  float b_cost = 0.0, a_cost = 0.0;
  std::vector< bool > active(f_units, false);

  for(int i = init_reference_time; i < reference_time; i++) {
    for(int j = 0; j < s->omega; j++) {
      if(s->s[i][j].first != NULL) {
	//Add the base cost of target instruction
	TargetInstruction *t_inst = s->s[i][j].second.first;
	b_cost += t_inst->get_base_cost();
	std::vector< int > use_fu = t_inst->get_use_funits();
	for(int k = 0; k < use_fu.size(); k++) {
	  assert(use_fu[k] < f_units);
	  //if active alread true: specification issue!
	  int fu = use_fu[k];
	  if(active[fu]) {
	    std::cerr << "Warning: same FU (" << fu << ") used in the same time slot.\n";
	  } 
	  active[fu] = true;
	  if(!power_profile->activation_status[fu]) {
	    a_cost += _architecture->get_activation_cost(fu);
	    power_profile->activation_status[fu] = true;
	  }
	}
	power_profile->instructions[j] = t_inst;
      } else {
	power_profile->instructions[j] = NULL;
      }
    }
    for(int j = 0; j < f_units; j++) {
      if(active[j] != power_profile->activation_status[j]) {
	a_cost += _architecture->get_activation_cost(j);
	power_profile->activation_status[j] = active[j];
      }
    }
  }
  //The last instruction slot is not in the power profile, but we count e
  for(int i = 0; i < s->omega; i++) {
    if(s->s[reference_time][i].first != NULL) {
      //base cost
      TargetInstruction *t_inst = s->s[reference_time][i].second.first;
      b_cost += t_inst->get_base_cost();
      std::vector< int > use_fu = t_inst->get_use_funits();
      for(int j = 0; j < use_fu.size(); j++) {
	int fu = use_fu[j];
	assert(fu < f_units);
	if(!power_profile->activation_status[fu]) {
	  a_cost += _architecture->get_activation_cost(fu);
	}
      }
    }
  }

  e += (a_cost + b_cost);
  return std::make_pair(true, e);
}

Schedule *ENode::get_schedule(void)
{
  return s;
}

NodeAttr *ENode::get_node_attr(void) const
{
  return node_attr;
}

float ENode::get_e(void)
{
  return e;
}

int ENode::get_tau(void) const
{
  return reference_time + time_profile->get_depth();
}
