#include <iterator>
#include <algorithm>
#include "tp.hh"

#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b)) 

const size_t TimeProfile::primes[MP][MP] = {
     3,    5,    7,   11,   13,   17,   19,   23,   29,   31,   37,   41,   43,   47,   53, 
    59,   61,   67,   71,   73,   79,   83,   89,   97,  101,  103,  107,  109,  113,  127, 
   131,  137,  139,  149,  151,  157,  163,  167,  173,  179,  181,  191,  193,  197,  199, 
   211,  223,  227,  229,  233,  239,  241,  251,  257,  263,  269,  271,  277,  281,  283, 
   293,  307,  311,  313,  317,  331,  337,  347,  349,  353,  359,  367,  373,  379,  383, 
   389,  397,  401,  409,  419,  421,  431,  433,  439,  443,  449,  457,  461,  463,  467, 
   479,  487,  491,  499,  503,  509,  521,  523,  541,  547,  557,  563,  569,  571,  577, 
   587,  593,  599,  601,  607,  613,  617,  619,  631,  641,  643,  647,  653,  659,  661, 
   673,  677,  683,  691,  701,  709,  719,  727,  733,  739,  743,  751,  757,  761,  769, 
   773,  787,  797,  809,  811,  821,  823,  827,  829,  839,  853,  857,  859,  863,  877, 
   881,  883,  887,  907,  911,  919,  929,  937,  941,  947,  953,  967,  971,  977,  983, 
   991,  997, 1009, 1013, 1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063, 1069, 1087, 
  1091, 1093, 1097, 1103, 1109, 1117, 1123, 1129, 1151, 1153, 1163, 1171, 1181, 1187, 1193, 
  1201, 1213, 1217, 1223, 1229, 1231, 1237, 1249, 1259, 1277, 1279, 1283, 1289, 1291, 1297, 
  1301, 1303, 1307, 1319, 1321, 1327, 1361, 1367, 1373, 1381, 1399, 1409, 1423, 1427, 1429, 
};

TimeProfile::TimeProfile(int _size) 
  : reserv_tab_t(_size), hash_value(0)
{
  nodes.clear();
  latency_map.clear();
}

TimeProfile::~TimeProfile()
{
}

TimeProfile::TimeProfile(const TimeProfile & _tp)
  : reserv_tab_t(_tp.size()),
    nodes(_tp.nodes), latency_map(_tp.latency_map)
{
  for(int i = 0; i < (*this).size(); i++) {
    (*this)[i] = _tp[i];
  }
  hash_value = _tp.hash_value;
}

bool TimeProfile::operator==(const TimeProfile & _time_profile) const
{
  if(hash_value != _time_profile.hash_value) {
    return false;
  }
  if(nodes != _time_profile.nodes) {
    return false;
  }
  if(latency_map != _time_profile.latency_map) {
    return false;
  }
  int size = this->size();
  for(int i = 0; i < size; i++) {
    int v1_length = (*this)[i].size(), v2_length = _time_profile[i].size();
    if(v1_length != v2_length) {
      return false;
    }
    for(int j = 0; j < v1_length; j++) {
      if((*this)[i][j] != _time_profile[i][j]) {
	return false;
      }
    }
  }
  return true;
}

int TimeProfile::insert(const vertex_set_t & _nodes,
			const reserv_tab_t & _reserv_tab,
			DAG & _dag)
{

  assert(_reserv_tab.size() == (*this).size());
	
  //build the set of predecessors to selected nodes
  //if they are in the TP, eject them by inserting NOPs
  predecessor_iterator vi, vi_end;
  int offset = 0;
  for(vertex_set_t::const_iterator ui = _nodes.begin(); ui != _nodes.end(); ++ui) {
    vertex_id u_id = *ui;
    for(tie(vi, vi_end) = predecessors(u_id, _dag); vi != vi_end; ++vi) {
      vertex_id v_id = *vi;
      if(nodes.find(*vi) != nodes.end()) {
	offset = max(offset, latency_map[(*vi)]);
      }
    }
  }
  //advance in time profile
  if(offset > 0) { 
    advance(offset); 
  }
	
  //latency of the instruction, length of reservation table
  int latency = 0;

 _recheck:	
  for(int i = 0; i < _reserv_tab.size(); i++) {
    latency = max(latency, _reserv_tab[i].size());
    int check_size = min((*this)[i].size(), _reserv_tab[i].size());
		
    for(int j = 0; j < check_size; j++) {
      if((_reserv_tab[i][j] == BUSY) && ((*this)[i][j] == BUSY)) {
	advance(1);
	offset++;
	goto _recheck;
      }
    }
  }
  //We can insert now. The latency of the instruction is latency.
  insert_reserv_tab(_nodes, _reserv_tab, latency);
  //Comput hash_value
  compute_hash();
	
  return offset;
}

int TimeProfile::get_depth(void)
{
  int depth = 0, size = (*this).size();
  for(int i = 0; i < size; i++) {
    depth = max(depth, ((*this)[i]).size());
  }
  return depth;
}

///Advance a given numbe of clock cycles
void TimeProfile::advance(int _offset)
{
  for(int i = 0; i < (*this).size(); i++) {
    for(int j = 0; j < _offset; j++) {
      if((*this)[i].size()) {
	(*this)[i].erase((*this)[i].begin());
      }
    } 
  }
  //Update nodes in the TimeProfile
  for(vertex_set_t::iterator vi = nodes.begin();
      vi != nodes.end(); ++vi) {
    if((latency_map[*vi] -= _offset) <= 0) {
      latency_map.erase(*vi);
      nodes.erase(vi);
    }
  }
}

void TimeProfile::insert_reserv_tab(const vertex_set_t & _nodes,
				    const reserv_tab_t & _reserv_tab, int _latency)
{
  for(int i = 0; i < _reserv_tab.size(); i++) {
    int tp_entry_size = (*this)[i].size();
    int rt_entry_size = _reserv_tab[i].size();
    if(tp_entry_size < rt_entry_size) {
      //Need new entries in TimeProfile vector
      for(int j = 0; j < tp_entry_size; j++) {
	if(_reserv_tab[i][j] == BUSY) {
	  assert((*this)[i][j] == EMPTY);
	  (*this)[i][j] = BUSY;
	}
      }
      for(int j = tp_entry_size; j < rt_entry_size; j++) {
	(*this)[i].push_back(_reserv_tab[i][j]);
      }
      assert((*this)[i].size() == rt_entry_size);
    } else {
      for(int j = 0; j < rt_entry_size; j++) {
	if(_reserv_tab[i][j] == BUSY) {
	  assert((*this)[i][j] == EMPTY);
	  (*this)[i][j] = BUSY;
	}
      }
    }
  }
  //Insert nodes and their latency (each node in the pattern has the same
  //latency).
  for(vertex_set_t::const_iterator vi = _nodes.begin(); vi != _nodes.end();
      ++vi) {

    /*//DEBUG
      std::cout << "vi: " << *vi << std::endl;
      copy(nodes.begin(), nodes.end(), 
      std::ostream_iterator< vertex_id >(std::cout, " "));
      std::cout << std::endl;
    */
    if(_latency > 0) {
      assert(nodes.find(*vi) == nodes.end());
      assert(latency_map.find(*vi) == latency_map.end());
      nodes.insert(*vi);
      latency_map[*vi] = _latency;
    }
  }
}

void TimeProfile::compute_hash(void)
{
  hash_value = 0;

  //Time profile shape
  for(int i = 0; i < (*this).size(); i++) {
    for(int j = 0; j < (*this)[i].size(); j++) {
      hash_value += primes[i % MP][j % MP];
    }
  }
  //Nodes
  size_t nodes_hash = 1;
  for(vertex_set_t::const_iterator vi = nodes.begin(); vi != nodes.end();
      ++vi) {
    nodes_hash *= (size_t)(*vi);
  }
  hash_value += nodes_hash;
}

/*
  TimeProfile & TimeProfile::operator=(const TimeProfile & _tp)
  //	: nodes(_tp.nodes), latency_map(_tp.latency_map) //, std::vector< bit_vector_t >(_tp.size())
  {
  nodes = _tp.nodes;
  latency_map = _tp.latency_map;
  for(int i = 0; i < _tp.size(); i++) {
  for(int j = 0; j < _tp[i].size(); j++) {
  (*this)[i].push_back(_tp[i][j]);
  }
  }
  return *this;
  }
*/

/*
  FUNIT TimeProfile::fu_map;

  Slot::Slot()
  {
  nodes.clear();
  l = o = 0;
  }

  Slot & Slot::operator=(const Slot & slot)
  {
  this->o = slot.o;
  this->l = slot.l;
  NODES::const_iterator si = slot.nodes.begin();
  for(; si != slot.nodes.end(); ++si) {
  this->nodes.insert(this->nodes.end(), (*si));
  }
  return *this;
  }

  Slot::Slot(const Slot & slot)
  {
  this->o = slot.o;
  this->l = slot.l;
  NODES::const_iterator si = slot.nodes.begin();
  for(; si != slot.nodes.end(); ++si) {
  this->nodes.insert(this->nodes.end(), (*si));
  }
  }

  //Add a functional unit to the system: max latency and max occupation for that unit.
  void TimeProfile::add_functional_unit(char *name, int latency, int occupation)
  {
  assert(fu_map.find(name) == fu_map.end());
  fu_map[name].l = latency;
  fu_map[name].o = occupation;
  (*this).push_back(new UnitTimeProfile(latency));
  fu_map[name].pos = (*this).size() - 1;
  }

  TimeProfile::TimeProfile(const TimeProfile & tp)
  {
  int f = (tp.size());
  for(int i = 0; i < f; i ++) {
  int size = (*(tp[i])).size();
  (*this).push_back(new UnitTimeProfile(size));
  //Copy all elements
  (*(*this)[i]) = *(tp[i]);
  }
  }

  TimeProfile::~TimeProfile()
  {
  int f = (*this).size();
  for(int i = 0; i < f; i++) {
  assert((*this)[i]);
  delete (*this)[i];
  }
  }

  std::ostream & operator<<(std::ostream & out, const Slot & slot) {
  out << "l:" << slot.l << " o:" << slot.o;
  return out;
  }

  std::ostream & operator<<(std::ostream & out,
  const UnitTimeProfile & unit_time_profile) {
  int l = unit_time_profile.size();
  for(int i = 0; i < l; i++) {
  out << "|" << unit_time_profile[i] << "|";
  }
  return out;
  }
*/

//Debug
std::ostream & operator<<(std::ostream & _out, const TimeProfile & _time_profile) {
  int f = _time_profile.size();
  _out << "nodes: { ";
  std::copy(_time_profile.nodes.begin(), _time_profile.nodes.end(), 
      std::ostream_iterator< vertex_id >(_out, " "));
  _out << "}\n";
  _out << "latencies: (v, l): { ";

  std::map< vertex_id, int > latencies = _time_profile.latency_map;
  std::map< vertex_id, int >::iterator li = latencies.begin(),
    li_end = latencies.end();
  for(; li != li_end; ++li) {
    _out << "(" << (*li).first << ", " << (*li).second << ") ";
  }

  _out << "}\ntime profile:\n";
  for(int i = 0; i < f; i++) {
    _out << i << ": ";
    int size = _time_profile[i].size();
    for(int j = 0; j < size; j++) {
      _out << _time_profile[i][j] << " ";
    }
    _out << std::endl;
  }
  return _out;
}
